@@ -402,6 +402,21 @@ def test_ls_on_file(tmpdir):
     assert fs.exists(resource)
     assert fs.ls(tmpdir) == fs.ls(resource)
     assert fs.ls(resource, detail=True)[0] == fs.info(resource)
+    assert fs.ls(resource, detail=False)[0] == resource
+
+
+def test_ls_on_files(tmpdir):
+    tmpdir = make_path_posix(str(tmpdir))
+    fs = LocalFileSystem()
+    resources = [f"{tmpdir}/{file}" for file in ["file_1.json", "file_2.json"]]
+    for r in resources:
+        fs.touch(r)
+
+    actual_files = fs.ls(tmpdir, detail=True)
+    assert {f["name"] for f in actual_files} == set(resources)
+
+    actual_files = fs.ls(tmpdir, detail=False)
+    assert set(actual_files) == set(resources)
 
 
 @pytest.mark.parametrize("file_protocol", ["", "file://"])
